..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _ec2api-common:

.. list-table:: Description of Common Service configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description

   * - **[DEFAULT]**
     -

   * - ``api_paste_config`` = ``api-paste.ini``

     - (String) File name for the paste.deploy config for ec2api

   * - ``max_header_line`` = ``16384``

     - (Integer) Maximum line size of message headers to be accepted. max_header_line may need to be increased when using large tokens (typically those generated by the Keystone v3 API with big service catalogs).

   * - ``tcp_keepidle`` = ``600``

     - (Integer) Sets the value of TCP_KEEPIDLE in seconds for each server socket. Not supported on OS X.

   * - ``wsgi_default_pool_size`` = ``1000``

     - (Integer) Size of the pool of greenthreads used by wsgi

   * - ``wsgi_log_format`` = ``%(client_ip)s "%(request_line)s" status: %(status_code)s len: %(body_length)s time: %(wall_seconds).7f``

     - (String) A python format string that is used as the template to generate log lines. The following values can be formatted into it: client_ip, date_time, request_line, status_code, body_length, wall_seconds.
