# Copyright 2014
# The Cloudscaling Group, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:mod:`ec2api` -- Cloud IaaS Platform
===================================

.. automodule:: ec2api
   :platform: Unix
   :synopsis: Infrastructure-as-a-Service Cloud platform.
"""

import six

import gettext


if six.PY2:
    gettext.install('ec2api', unicode=1)
else:
    gettext.install('ec2api')
