..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _ec2api-s3:

.. list-table:: Description of S3 Client configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description

   * - **[DEFAULT]**
     -

   * - ``image_decryption_dir`` = ``/tmp``

     - (String) Parent directory for tempdir used for image decryption

   * - ``s3_region`` = ``RegionOne``

     - (String) Region of S3 server

   * - ``s3_url`` = ``http://$my_ip:3334``

     - (String) URL to S3 server

   * - ``x509_root_private_key`` = ``None``

     - (String) Path to ca private key file
