..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _ec2api-service:

.. list-table:: Description of EC2API Service configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description

   * - **[DEFAULT]**
     -

   * - ``ec2api_listen`` = ``0.0.0.0``

     - (String) The IP address on which the EC2 API will listen.

   * - ``ec2api_listen_port`` = ``8788``

     - (Integer) The port on which the EC2 API will listen.

   * - ``ec2api_use_ssl`` = ``False``

     - (Boolean) Enable ssl connections or not for EC2 API

   * - ``ec2api_workers`` = ``None``

     - (Integer) Number of workers for EC2 API service. The default will be equal to the number of CPUs available.

   * - ``ssl_cert_file`` = ``None``

     - (String) SSL certificate of API server

   * - ``ssl_key_file`` = ``None``

     - (String) SSL private key of API server

   * - ``use_forwarded_for`` = ``False``

     - (Boolean) Treat X-Forwarded-For as the canonical remote address. Only enable this if you have a sanitizing proxy.
