..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _ec2api-metadata:

.. list-table:: Description of Metadata configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description

   * - **[DEFAULT]**
     -

   * - ``metadata_listen`` = ``0.0.0.0``

     - (String) The IP address on which the metadata API will listen.

   * - ``metadata_listen_port`` = ``8789``

     - (Integer) The port on which the metadata API will listen.

   * - ``metadata_use_ssl`` = ``False``

     - (Boolean) Enable ssl connections or not for EC2 API Metadata

   * - ``metadata_workers`` = ``None``

     - (Integer) Number of workers for metadata service. The default will be the number of CPUs available.

   * - **[metadata]**
     -

   * - ``auth_ca_cert`` = ``None``

     - (String) Certificate Authority public key (CA cert) file for ssl

   * - ``cache_expiration`` = ``15``

     - (Integer) This option is the time (in seconds) to cache metadata. Increasing this setting should improve response times of the metadata API when under heavy load. Higher values may increase memory usage, and result in longer times for host metadata changes to take effect.

   * - ``metadata_proxy_shared_secret`` =

     - (String) Shared secret to sign instance-id request

   * - ``nova_client_cert`` =

     - (String) Client certificate for nova metadata api server.

   * - ``nova_client_priv_key`` =

     - (String) Private key of client certificate.

   * - ``nova_metadata_insecure`` = ``False``

     - (Boolean) Allow to perform insecure SSL (https) requests to nova metadata

   * - ``nova_metadata_ip`` = ``127.0.0.1``

     - (String) IP address used by Nova metadata server.

   * - ``nova_metadata_port`` = ``8775``

     - (Integer) TCP Port used by Nova metadata server.

   * - ``nova_metadata_protocol`` = ``http``

     - (String) Protocol to access nova metadata, http or https
